/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"

const char* get_char_ip(int type, struct proxyinfo *pi) {
	unsigned long int ip = get_uint_ip(type, pi);
	struct in_addr addr;
	addr.s_addr = ip;

	return inet_ntoa(addr);
}

unsigned long int get_uint_ip(int type, struct proxyinfo* pi) {
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	unsigned long int *ip_ptr;
	int *fd_ptr;

	if (type == GET_IP_SERVER) {
		ip_ptr = &si->connection.server_ip;
		fd_ptr = &si->connection.serversocket;
	} else if (type == GET_IP_CLIENT) {
		ip_ptr = &si->connection.client_ip;
		fd_ptr = &si->connection.clientsocket;
	} else {
		return UINT_MAX;
	}

	/* The IP is already known, return it */
	if (*ip_ptr != UINT_MAX) {
		return *ip_ptr;
	}

	/* The IP is not known but if we are connected we look at the socket */
	if (*fd_ptr != -1) {
		*ip_ptr = get_uint_peer_ip(*fd_ptr);
		return *ip_ptr;
	}

	/* if we are testing for the destination, i.e. SERVER and
	 * clntinfo->destination happens to be an IP, use this one */

	if (type == GET_IP_SERVER && ci->destination) {
		*ip_ptr = inet_addr(ci->destination);
	}

	return *ip_ptr;
}


/* parsesock parses a comma separated list of IP and Port like in
 *  PORT 127,0,0,1,15,216
 * or the PASV answer.
 */

int parsesock(const char* buffer, struct sockaddr_in *sin, int mode) {
	int i1, i2, i3, i4, lo, hi, count;
	unsigned long int iaddr;
	unsigned int port;
	char ipbuf[16];
	int consumed;
	const char* p;

	memset((void*)sin, 0, sizeof(*sin));
	if (mode == ACTIVE) {
		count = sscanf(buffer, "%d,%d,%d,%d,%d,%d%n",
				&i1, &i2, &i3, &i4, &hi, &lo, &consumed);
	} else if (mode == PASSIVE) {
		int accepted = 0;
		int deny = 0;
		count = sscanf(buffer, "%d,%d,%d,%d,%d,%d%n",
				&i1, &i2, &i3, &i4, &hi, &lo, &consumed);
		p = buffer + consumed;
		while (*p && accepted < 10) {
			switch(*p) {
				case '.':
				case ';':
				case ')':
				case '\r':
				case '\n':
					accepted++;
					consumed++;
					break;
				default:
					deny = 1;
					break;
			}
			p++;
		}
		if (deny || accepted >= 10) {
			jlog(3, "Invalid address in the PASV or PORT command: "
					"%s", buffer);
			return -1;
		}
	} else {
		jlog(3, "Invalid mode in parsesock for buffer: %s", buffer);
		return -1;
	}
	/* sscanf must have read 6 arguments and all the parameters must be
	 * less than 255 ( 0xff ) and > 0
	 */
	if (!(count != 6 || i1 > 0xff || i2 > 0xff || i3 > 0xff || i4 > 0xff
			|| hi > 0xff || lo > 0xff
			|| i1 < 0 || i2 < 0 || i3 < 0 || i4 < 0
			|| hi < 0 || lo < 0
			/* there are even more characters than the ones
			 * we have parsed */
			|| consumed != strlen(buffer))) {
		snprintf(ipbuf, 16, "%d.%d.%d.%d",
				i1, i2, i3, i4);
		iaddr = inet_addr(ipbuf);
		if (iaddr != -1 || !strcmp(ipbuf, BROADCAST)) {
			sin->sin_family = AF_INET;
			sin->sin_addr.s_addr = iaddr;
			port = hi * 256 + lo;
			sin->sin_port = htons(port);
			return 0;
		}
		else {
			jlog(3, "Invalid address in the PASV or PORT command: %s",
					buffer);
		}
	}
	jlog(3, "Error parsing IP and port from %s", buffer);
	return -1;
}


