/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

/*
How does it work ?

we need to do a few basic tasks:

- check if we have the file.
- if we have not, hook into the transfer that will follow and write the file to
  disk
- if we have, serve the file from disk
- what about expiry?
  - expire according to date
  - according to a given size

*/

#include <sys/stat.h>
#include <fcntl.h>
#include <utime.h>
#include "../ftp_main.h"

/* only the user should be able to read/write the cache */
int cache_perms = S_IRWXU;
static int cache_available(struct cache_filestruct, struct proxyinfo* pi);
static int cache_want(struct cache_filestruct);
static int cache_add(struct cache_filestruct cfs, struct proxyinfo* pi);
static int cache_delete(struct cache_filestruct cfs, struct proxyinfo* pi, int);
static int cache_writefd(struct cache_filestruct cfs, struct proxyinfo* pi);
static char* cache_qualifypath(const struct cache_filestruct,
							struct proxyinfo*);
static char* cache_qualifyfile(const struct cache_filestruct,
							struct proxyinfo*);
static int recursive_mkdir(const char* pathname, int perms);
static int filter_cache_write(struct transfer_t*, struct proxyinfo*,
							struct filter_t*);
static void cache_filestruct_free(struct cache_filestruct cfs);

char* getftpwd(struct proxyinfo*);
static unsigned long int getftpsize(char* filename, struct proxyinfo*);
static time_t getftpmdtm(const char* filename, struct proxyinfo*);

/* for the filter */
static struct cache_filestruct filter_get_cfs;
static int cache_transfer_complete;

struct config_directives_list diskcache_configuration_directives_array[] = {
	{"enabled",	TAG_ALL, "no", EM, WSP },
	{"directory",	TAG_ALL, "/tmp", EM, WSP },
	{"minfilesize",	TAG_ALL, "200", EM, WSP },
	{"maxfilesize",	TAG_ALL, "1000M", EM, WSP },
	{ (char*) 0,	0, (char*) 0, 0, 0 }
};

struct config_namespace_directives_list diskcache_configuration_directives = {
	"diskcache",		/* namespace name */
	diskcache_configuration_directives_array
};

struct config_option_values_list diskcache_configuration_options_array[] = {
	{"enabled",	{ TRUEFALSE, TERM } },
	{ TERM,         { TERM } }
};

struct config_namespace_options_list diskcache_configuration_options = {
	"diskcache",
	diskcache_configuration_options_array
};

int diskcache_add_configuration() {
	config_configuration_list_add(&diskcache_configuration_directives);
	config_options_list_add(&diskcache_configuration_options);
	return 0;
}

int diskcache_config_init() {
	config_register_hook(0, diskcache_add_configuration);
	return 0;
}

int diskcache_commandhandler_init() {
	transfermode_register(200,
			TRANSFERMODE_RECV_SERVER,
			"diskcache",
			transfermode_diskcache_in,
			transfermode_diskcache_in_from_cache_init,
			transfermode_diskcache_deinit_server,
			NULL,
			"Diskcache IN from cache (server)");

	filter_register("diskcachewrite", FILTER_TYPE_DOWNSTREAM,
			filter_cache_write,
			0, /* not active by default */
			"write files to the cache");

	return 0;
}

int diskcache_init() {
	module_hook_register(MODULE_HOOK_PROXY_INIT_CONFIG,
			     diskcache_config_init,
			     "diskcache config init");

	module_hook_register(MODULE_HOOK_FTP_CMDHANDLER_INIT,
			     diskcache_commandhandler_init,
			     "diskcache commandhandler init");
	return 0;
}

static int filter_cache_write(struct transfer_t* t,
				struct proxyinfo* pi,
				struct filter_t* __unused__) {

	static int doing;
	static int fd;

	ASSUME(t != NULL);

	if (doing == 0) {
		/* start of a transfer */
		fd = cache_writefd(filter_get_cfs, pi);
		doing = 1;
	}
	if (writeall(fd, t->buffer, t->bufsize) < 0) {
		close(fd);
		filter_deactivate("diskcachewrite", FILTER_TYPE_DOWNSTREAM);
		doing = 0;
		return -1;
	}
	if (t->status == TRANSFER_STATUS_ERROR
				|| t->status == TRANSFER_STATUS_FINISHED) {
		close(fd);
		filter_deactivate("diskcachewrite", FILTER_TYPE_DOWNSTREAM);
		if (t->status == TRANSFER_STATUS_FINISHED) {
			if (cache_add(filter_get_cfs, pi) < 0) {
				/* don't care if the file is there or not */
				cache_delete(filter_get_cfs, pi, 0);
			}
		} else {
			cache_delete(filter_get_cfs, pi, 1);
		}
		cache_filestruct_free(filter_get_cfs);
		doing = 0;
	}
	return 0;
}

static
void cache_filestruct_copy(const struct cache_filestruct src,
		           struct cache_filestruct *dst) {

	ASSUME(dst != NULL);

	dst->host = strnulldup(src.host);
	dst->port = src.port;
	dst->user = strnulldup(src.user);
	dst->filepath = strnulldup(src.filepath);
	dst->filename = strnulldup(src.filename);
	dst->size = src.size;
	dst->date = src.date;
}

static
void cache_filestruct_free(struct cache_filestruct cfs) {
	free_and_zero_char(&cfs.host);
	free_and_zero_char(&cfs.user);
	free_and_zero_char(&cfs.filepath);
	free_and_zero_char(&cfs.filename);
}

static
struct cache_filestruct cache_gather_info(const char* filename,
						struct proxyinfo* pi) {
	struct clientinfo* ci = pi->ci;

	char* pwd;
	char* complete_fname;
	size_t size;
	struct cache_filestruct cfs;

	if (filename[0] != '/') {
		/* get the directory */
		pwd = getftpwd(pi);
	} else {
		pwd = strdup("");
		enough_mem(pwd);
	}
	size = strlen(filename) + 1 + strlen(pwd) + 1;
	complete_fname = (char*) malloc(size);
	enough_mem(complete_fname);
	complete_fname = rel2abs(filename, pwd, complete_fname, size);
	if (!complete_fname) {
		jlog(4, "Error in rel2abs: filename: %s, path: %s",
							filename, pwd);
	}
	free(pwd);
	cfs.filepath = extract_path(complete_fname);
	cfs.filename = extract_file(complete_fname);
	cfs.size = getftpsize(complete_fname, pi);
	cfs.date = getftpmdtm(complete_fname, pi);
	free(complete_fname);

	cfs.user = ci->user;
	cfs.host = ci->destination;
	cfs.port = ci->destinationport;

	return cfs;
}

static
int cache_available(struct cache_filestruct cfs, struct proxyinfo *pi) {
	char* fname = cache_qualifyfile(cfs, pi);
	struct stat st;

	if (stat(fname, &st) < 0) {
		if (errno == ENOENT) {
			return CACHE_NOTAVL_EXIST;
		}
		/* other error */
		jlog(2, "Could not stat %s: %s", fname, strerror(errno));
		return -1;
	}

	if (cfs.size != st.st_size) {
		/* the new file seems to differ in size -> delete our copy */
		jlog(8, "cache copy differs in size");
		cache_delete(cfs, pi, 1);
		return CACHE_NOTAVL_SIZE;
	}

	if (cfs.date != st.st_mtime) {
		/* the new file seems to differ in the date -> delete our
		 * copy */
		jlog(8, "cache copy differs in the date");
		cache_delete(cfs, pi, 1);
		return CACHE_NOTAVL_DATE;
	}

	return CACHE_AVAILABLE;
}

static
int cache_add(struct cache_filestruct cfs, struct proxyinfo* pi) {
	struct stat st;
	char* fname = cache_qualifyfile(cfs, pi);
	struct utimbuf ut;

	/* okay, we're sure the path exists, let's try to create a file */

	/* compare size, set date... */
	if (stat(fname, &st) < 0) {
		jlog(2, "Could not stat %s: %s", fname, strerror(errno));
		return -1;
	}
	if (st.st_size != cfs.size) {
		jlog(6, "Had to delete %s again. Size did not match", fname);
		return -1;
	}

	/* set the date */
	ut.actime = ut.modtime = cfs.date;
	if (utime(fname, &ut) < 0) {
		jlog(6, "Could net set date/time information to %s: %s",
				fname, strerror(errno));
		return -1;
	}
	return 0;
}

static
int cache_writefd(struct cache_filestruct cfs, struct proxyinfo* pi) {
/*      Add a file to the cache.
	The directory structure is like
	<cache_directory> / <user>@<host>:<port> / <filepath> / <filename>
*/
	char* path;
	char* filefn;
	int fd;

	if (!cache_want(cfs)) {
		return -1;
	}

	/* just delete for sure, don't care about the return value */
	cache_delete(cfs, pi, 0);

	path = cache_qualifypath(cfs, pi);
	ASSUME(path != NULL);
	if (recursive_mkdir(path, cache_perms) < 0 && errno != EEXIST) {
		jlog(2, "Could not create directory %s: %s",
			path, strerror(errno));
		return -1;
	}

	filefn = cache_qualifyfile(cfs, pi);
	fd = creat(filefn, cache_perms);
	if (fd < 0) {
		jlog(2, "Could not create data file %s in cache: %s",
				filefn, strerror(errno));
	}
	return fd;
}

static
int cache_delete(struct cache_filestruct cfs, struct proxyinfo* pi, int warn) {
	char* datafile;
	int err = 0;

	datafile = cache_qualifyfile(cfs, pi);

	if (unlink(datafile) < 0 && warn) {
		jlog(2, "Could not unlink file %s: %s",
				datafile, strerror(errno));
		err = -1;
	}
	return err;
}

static
int cache_readfd(struct cache_filestruct cfs,
		 struct proxyinfo* pi) {
	char* fname = cache_qualifyfile(cfs, pi);
	int fd;

	if (cache_available(cfs, pi) != CACHE_AVAILABLE) {
		return -1;
	}

	fd = open(fname, O_RDONLY);
	return fd;
}

static
int cache_want(struct cache_filestruct cfs) {
/*
	See if we want a file to be added to the cache.
*/
	unsigned long int minfilesize
			= config_get_size("diskcache.minfilesize", 0);
	unsigned long int maxfilesize
			= config_get_size("diskcache.maxfilesize", ULONG_MAX);
	return (cfs.size <= maxfilesize && cfs.size >= minfilesize);
}

static
char* cache_qualifypath(const struct cache_filestruct cfs,
			struct proxyinfo* pi) {
	size_t size;
	static char* path;
	const char* hostname;
	unsigned long iaddr = pi->si->connection.server_ip;
	const char* cache_directory = config_get_option("diskcache.directory");

	ASSUME(cache_directory != NULL);
	ASSUME(cfs.host != NULL);

	/* try to look up the server ip - XXX should use server_name first
	 * maybe */
	if (!(hostname = hostent_get_name(&pi->si->hostcache, iaddr))) {
		hostname = cfs.host;
	}
	/* hostname is either the name of the host or the IP converted to a
	 * char* */
	ASSUME(hostname != NULL);

	size =   strlen(cache_directory) + 1
	       + strlen(cfs.user)     + 1
	       + strlen(hostname)     + 1
	       + 20 /* port */
	       + strlen(cfs.filepath) + 1
	       + 1;

	if (path) {
		path = (char*) realloc(path, size);
	} else {
		path = (char*) malloc(size);
	}
	enough_mem(path);

	snprintf(path, size, "%s/%s@%s:%d/%s",
			cache_directory,
			cfs.user,
			hostname,
			cfs.port,
			cfs.filepath);
	char_squeeze(path, '/');

	return path;
}

static
char* cache_qualifyfile(const struct cache_filestruct cfs,
						struct proxyinfo* pi) {
	char* path;
	size_t size;
	static char* filename;

	path = cache_qualifypath(cfs, pi);
	ASSUME(path != NULL);
	size = strlen(path) + 1 + strlen(cfs.filename) + 1;

	if (filename) {
		filename = (char*) realloc(filename, size);
	} else {
		filename = (char*) malloc(size);
	}
	enough_mem(filename);

	snprintf(filename, size, "%s/%s", path, cfs.filename);
	char_squeeze(filename, '/');

	return filename;
}

/* XXX compare with courier... */
int recursive_mkdir(const char* pathname, int perms) {
	char* tocreate;
	int sidx = 0;
	int failed = 0;

	tocreate = (char*) malloc(strlen(pathname) + 1);
	enough_mem(tocreate);

	do {
		while(pathname[sidx] && pathname[sidx] != '/') {
			tocreate[sidx] = pathname[sidx];
			sidx++;
		}
		tocreate[sidx] = pathname[sidx];
		sidx++;
		tocreate[sidx] = '\0';

		jlog(9, "Creating %s\n", tocreate);
		if (mkdir(tocreate, perms) < 0 && errno != EEXIST) {
			failed = -1;
		}
	} while(pathname[sidx] && failed == 0);

	free(tocreate);
	return -failed;
}

/* not static - this is also used by ftp_squidcache.c */
char* getftpwd(struct proxyinfo *pi) {
	int ss = pi->si->connection.serversocket;
	struct ftp_message msg;

	char* dir;
	char* dirstart, *dirend;
	size_t dirsize;

	say(ss, "PWD\r\n");
	msg = ftp_readall(ss);
	/*  257 "/home/joe" is current directory.  */
	/*  257 "/home/joe/hi""ho lo di'da" is current directory. */

	if (msg.error) {
		jlog(4, "PWD failed!");
		return (char*) 0;
	}
	if (msg.code != 257) {
		jlog(4, "PWD failed: %s", msg.msg);
		ftp_message_free(&msg);
		return (char*) 0;
	}
	dirstart = strchr(msg.msg, '"');
	dirend = strrchr(msg.msg, '"');

	if (!dirstart || !dirend || dirstart == dirend) {
		jlog(4, "Could not parse PWD response: %s", msg.msg);
		ftp_message_free(&msg);
		return (char*) 0;
	}

	dirstart++;
	dirend--;

	dirsize = dirend - dirstart + 1 + 1;
	ASSUME(dirsize >= 2); /* at least / + term */

	dir = (char*) malloc(dirsize);
	enough_mem(dir);

	snprintf(dir, dirsize, "%s", dirstart);
	ftp_message_free(&msg);
	char_squeeze(dir, '"');

	return dir;
}

static
time_t getftpmdtm(const char* filename, struct proxyinfo *pi) {
/*
 *	ftp> quote mdtm bla
 *	213 20010224102705
 */
	int ss = pi->si->connection.serversocket;

	size_t cmdsize = strlen("MDTM ") + strlen(filename) + 2 + 1;
	struct tm tms;
	int i;
	struct ftp_message msg;
	char* cmd = (char*) malloc(cmdsize);

	enough_mem(cmd);
	snprintf(cmd, cmdsize, "MDTM %s\r\n", filename);

	memset(&tms, 0, sizeof(tms));

	say(ss, cmd);
	free(cmd);
	msg = ftp_readall(ss);
	if (msg.error) {
		jlog(4, "Error reading MDTM answer!");
		return (time_t)(-1);
	}
	if (msg.code != 213) {
		jlog(4, "Error reading MDTM answer: %s", msg.msg);
		ftp_message_free(&msg);
		return (time_t)(-1);
	}

	/* 213 20010224102705 */
	i = sscanf(msg.msg, "213 %4d%2d%2d%2d%2d%2d",
			&tms.tm_year, &tms.tm_mon, &tms.tm_mday,
			&tms.tm_hour, &tms.tm_min, &tms.tm_sec);

	if (i != 6) {
		jlog(4, "Error parsing MDTM answer: %s", msg.msg);
		ftp_message_free(&msg);
		return (time_t)(-1);
	}
	tms.tm_year -= 1900;  /* tm_year contains the number of years */
			      /* since 1900 */
	tms.tm_mon -= 1;      /* tm_mon starts with 0 */

	ftp_message_free(&msg);
	return mktime(&tms);
}

static
unsigned long int getftpsize(char* filename, struct proxyinfo *pi) {
/*
 * ftp> quote size speak.ps
 * 213 146617
 */
	int ss = pi->si->connection.serversocket;

	size_t cmdsize = strlen("SIZE ") + strlen(filename) + 2 + 1;
	unsigned long int size;
	int i;
	struct ftp_message msg;
	char* cmd = (char*) malloc(cmdsize);

	enough_mem(cmd);
	snprintf(cmd, cmdsize, "SIZE %s\r\n", filename);

	say(ss, cmd);
	free(cmd);
	msg = ftp_readall(ss);
	if (msg.error) {
		jlog(4, "Error reading SIZE answer!");
		return 0;
	}
	if (msg.code != 213) {
		jlog(4, "Error reading SIZE answer: %s", msg.msg);
		ftp_message_free(&msg);
		return 0;
	}
	i = sscanf(msg.msg, "213 %ld", &size);
	if (i != 1) {
		jlog(4, "Error parsing SIZE answer: %s", msg.msg);
		ftp_message_free(&msg);
		return 0;
	}
	ftp_message_free(&msg);
	return size;
}

int transfermode_diskcache_in_from_cache_init(const char* cmd,
					       const char** args,
					       struct proxyinfo* pi,
					       struct messagebuffer* msgbuf,
					       struct ftp_message* server_msg) {
	int cache_fd;
	struct cache_filestruct cfs;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);
	/* UNSPEC should not exist here because we replace it with a default
	 * during the login. */
	if (config_get_bool("diskcache.enabled", 0) == 0) {
		return TRANSFERMODE_NOTSUPPORTED;
	}

	if (strcmp(cmd, "RETR") != 0) {
		return TRANSFERMODE_NOTSUPPORTED;
	}
	cache_transfer_complete = 0;
	cfs = cache_gather_info(args[0], pi);
	cache_fd = cache_readfd(cfs, pi);
	if (cache_fd < 0) {
		if (cache_want(cfs)) {
			/* register for copying the file */
			cache_filestruct_copy(cfs, &filter_get_cfs);
			filter_activate("diskcachewrite",
						FILTER_TYPE_DOWNSTREAM);
		}
		return TRANSFERMODE_NOTSUPPORTED;
	} else {
		/* ok, we have the file here */
		/* XXX I don't like this... */
		pi->ci->data_connection.dataserversock = cache_readfd(cfs, pi);
	}

	return TRANSFERMODE_OK;
}


int transfermode_diskcache_in(int fd_in,
			   struct transfer_t* transfer,
			   struct proxyinfo* pi) {

	int ret;

	/* XXX there are log messages in transfermode_std_in containing the
	 * name "transfermode_std_in", this might be misleading... */
	jlog(9, "transfermode_diskcache_in() calls transfermode_std_in()");
	ret = transfermode_std_in(fd_in, transfer, pi);
	if (ret == 0 && transfer->status ==  TRANSFER_STATUS_FINISHED) {
		cache_transfer_complete = 1;
	}
	return ret;
}


int transfermode_diskcache_deinit_server(struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);

	pi->ci->data_connection.complete = cache_transfer_complete;

	if (cache_transfer_complete) {
		ftp_message_setm(server_msg,
					226, "Transfer complete",
					0, 0, -1);
	} else {
		/* XXX */
		ftp_message_setm(server_msg,
					226, "Transfer aborted",
					0, 0, -1);
	}
	
	close(pi->ci->data_connection.dataserversock);
	pi->ci->data_connection.dataserversock = -1;

	ASSUME(server_msg != NULL);
	ASSUME(server_msg->code);

	return TRANSFERMODE_OK;
}

