/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "../ftp_main.h"

/* XXX use the cache only in anonymous mode ? */

/* ----------- Squid cache ------------ */

static
int squid_fd;

struct config_directives_list squidcache_configuration_directives_array[] = {
	{"enabled",	TAG_ALL, "no", EM, WSP },
	{"proxyname",	TAG_ALL, (char*) 0, EM, WSP },
	/* what is the default port? 8080 or 3128 ? */
	{"proxyport",	TAG_ALL, "8080", EM, WSP },
	{ (char*) 0,	0, (char*) 0, 0, 0 }
};

struct config_namespace_directives_list squidcache_configuration_directives = {
	"squidcache",		/* namespace name */
	squidcache_configuration_directives_array
};

struct config_option_values_list squidcache_configuration_options_array[] = {
	{"enabled",	{ TRUEFALSE, TERM } },
	{ TERM,         { TERM } }
};

struct config_namespace_options_list squidcache_configuration_options = {
	"squidcache",
	squidcache_configuration_options_array
};

int squidcache_add_configuration() {
	config_configuration_list_add(&squidcache_configuration_directives);
	config_options_list_add(&squidcache_configuration_options);
	return 0;
}

int squidcache_config_init() {
	config_register_hook(0, squidcache_add_configuration);
	return 0;
}

int squidcache_commandhandler_init() {
	transfermode_register(200,
			TRANSFERMODE_RECV_SERVER,
			"squidcache",
			transfermode_squidcache_in,
			transfermode_squidcache_in_from_cache_init,
			transfermode_squidcache_deinit_server,
			NULL,
			"Squidcache IN from cache (server)");

	return 0;
}

int squidcache_init() {
	module_hook_register(MODULE_HOOK_PROXY_INIT_CONFIG,
			     squidcache_config_init,
			     "squidcache config init");

	module_hook_register(MODULE_HOOK_FTP_CMDHANDLER_INIT,
			     squidcache_commandhandler_init,
			     "squidcache commandhandler init");
	return 0;
}



static
int squid_connect(struct proxyinfo* pi,
		  struct ftp_message* server_msg) {

	const char* proxyname = config_get_option("squidcache.proxyname");
	/* the default should be set by the config */
	int proxyport = config_get_ioption("squidcache.proxyport", -1);


	/* however the user can still set a negative value for the port */
	if (proxyport < 0) {
		return TRANSFERMODE_NOTSUPPORTED;
	}
	if (!proxyname) {
		return TRANSFERMODE_NOTSUPPORTED;
	}

	squid_fd = openportname(proxyname, proxyport,
			INADDR_ANY, /* local address */
			(const struct portrangestruct*) 0,
			pi);

	if (squid_fd < 0) {
		/* XXX check code... */
		ftp_message_setm(server_msg,
				400, "Could not connect to cache proxy",
				ERROR_CONNECTIONFAILED,
				errno,
				-1);
		/* we return FAIL here because the user requested the cache. 
		 * so we just don't fall back without saying anything... */
		return TRANSFERMODE_FAIL;
	}

	return TRANSFERMODE_OK;
}

/* char* getftpwd(struct proxyinfo *pi) is defined in ftp_diskcache.c */
	
static
int squid_request(const char** args,
		  struct proxyinfo* pi,
		  struct ftp_message* server_msg) {

	/* GET ftp://user:pass@host:port/path/file HTTP/1.0 */

	struct clientinfo* ci = pi->ci;

	/* XXX which user should we take ? - make it configurable maybe? */
	const char* user = ci->user;
	const char* pass = ci->anon_user_pass;

	const char* host = ci->destination;
	int port = ci->destinationport;

	const char* file = args[0];
	char* path_m;

	size_t size, count;
	char* request;

	int success;
	char* line;

	if (port > 99999 || port < 0) {
		close(squid_fd);
		squid_fd = -1;
		/* XXX do we have to set server_msg ? */
		return TRANSFERMODE_FAIL;
	}

	/* XXX this won't work for all server transfermodes... */
	path_m = getftpwd(pi);

	if (!path_m || path_m[0] != '/') {
		jlog(2, "could not get the current working dir: %s",
				path_m ? path_m : "null");
		/* XXX check ftp response code... */
		ftp_message_setm(server_msg,
				400, "Could not get file from cache proxy",
				ERROR_CONNECTIONFAILED,
				errno,
				-1);

		free(path_m);
		close(squid_fd);
		squid_fd = -1;
		return TRANSFERMODE_FAIL;
	}

	size = strlen("GET ftp://")
		+ strlen(user)
		+ 1
		+ strlen(pass)
		+ 1
		+ strlen(host)
		+ 1
		+ 5 /* 65,000 */
		+ strlen(path_m)
		+ 1
		+ strlen(file)
		+ strlen(" HTTP/1.0\r\n\r\n")
		+ 1; /* term */

	request = (char*) malloc(size);
	snprintf(request, size, "GET ftp://%s:%s@%s:%d%s/%s HTTP/1.0\r\n\r\n",
				user, pass,
				host, port,
				path_m, file);
	free(path_m);

	/* we don't write the terminator... */
	count = writeall(squid_fd, request, size - 1);
	free(request);
	if (count < 0) {
		jlog(2, "could not write request to squid proxy: %s",
				strerror(errno));
		/* XXX check code... */
		ftp_message_setm(server_msg,
				400, "Could not get file from cache proxy",
				ERROR_CONNECTIONFAILED,
				errno,
				squid_fd);

		close(squid_fd);
		squid_fd = -1;

		/* we return FAIL here because the user requested the cache. 
		 * so we just don't fall back without saying anything... */
		return TRANSFERMODE_FAIL;
	}

	/* ok, now read the answer and parse it... */
	success = 0;
	do {
		line = readline(squid_fd);
		if (!line) {
			success = 0;
			break;
		}
		if (line[0] == '\0') {
			/* finished - this indicates an empty line,
			 * exactly what we're looking for  :-) */
			free(line);
			break;
		}
		if (checkbegin(line, "Content-Length")) {
			/* XXX */
		}
		if (checkbegin(line, "HTTP/1.0")) {
			if (strcmp(line, "HTTP/1.0 200 OK") == 0) {
				success = 1;
			}
		}
		if (checkbegin(line, "X-Cache: MISS ")) {
			jlog(9, "MISS on cache");
		}
		if (checkbegin(line, "X-Cache: HIT ")) {
			jlog(9, "HIT on cache");
		}
		free(line);
	} while (1);

	if (!success) {
		/* XXX check code... */
		ftp_message_setm(server_msg,
				400, "Got an error from the proxy",
				0, 0, squid_fd);

		close(squid_fd);
		squid_fd = -1;
		return TRANSFERMODE_FAIL;
	}

	ftp_message_setm(server_msg,
			125, "Opening connection for transfer",
			0, 0, squid_fd);

	return TRANSFERMODE_OK;
}

int transfermode_squidcache_in_from_cache_init(const char* cmd,
					       const char** args,
					       struct proxyinfo* pi,
					       struct messagebuffer* msgbuf,
					       struct ftp_message* server_msg) {
	int ret;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);
	/* UNSPEC should not exist here because we replace it with a default
	 * during the login. */

	if (strcmp(cmd, "RETR") != 0) {
		return TRANSFERMODE_NOTSUPPORTED;
	}
	ret = squid_connect(pi, server_msg);
	if (ret != TRANSFERMODE_OK) {
		return ret;
	}

	ret = squid_request(args, pi, server_msg);
	if (ret != TRANSFERMODE_OK) {
		return ret;
	}

	/* XXX I don't like this... */
	pi->ci->data_connection.dataserversock = squid_fd;

	return TRANSFERMODE_OK;
}

int
transfermode_squidcache_in(int fd_in,
			   struct transfer_t* transfer,
			   struct proxyinfo* pi) {

	/* the commandhandler opens up a file descriptor, passes it to us
	 * and we read the data from there. Then we do whatever is necessary
	 * to strip off the communication protocol and write chunks of data
	 * to *transfer. This is then passed through the filters and finally
	 * through the output transfermode. Then we are called again to read
	 * further */

	fd_set readset;
	struct timeval readtime;
	int ret;
	ssize_t count;

	/* SIG_BLOCK - SIG_UNBLOCK ? Or restart (test for EINTR?) */
	do {
		readtime.tv_sec = config_get_ioption("transfertimeout", 300);
		readtime.tv_usec = 0;

		FD_ZERO(&readset);
		FD_SET(fd_in, &readset);

		ret = select(fd_in + 1, &readset, NULL, NULL, &readtime);
		if (ret < 0 && errno == EINTR) {
			continue;
		}
		if (ret < 0) {
			jlog(2, "select error in transfermode_std_in: %s",
					strerror(errno));
			return -1;
		}
		if (ret == 0) {
			jlog(2, "select timeout in transfermode_std_in: %s",
					strerror(errno));
			return -1;
		}
	} while (ret < 0);

	if (!FD_ISSET(fd_in, &readset)) {
		/* shouldn't happen */
		return -1;
	}

	transfer->buffer = (char*) malloc(TRANSFER_BUFFERSIZE);
	enough_mem(transfer->buffer);
	count = read(fd_in, transfer->buffer, TRANSFER_BUFFERSIZE);
	if (count < 0) {
		jlog(2, "read() error in transfermode_std_in: %s",
				strerror(errno));
		free(transfer->buffer);
		return -1;
	}
	if (count == 0) {
		free(transfer->buffer);
		transfer->buffer = (char*) 0;
		transfer->bufsize = 0;
		transfer->status = TRANSFER_STATUS_FINISHED;
		return 0;
	}
	if (count < transfer->bufsize) {
		/* we didn't use the whole buffer */
		transfer->buffer = realloc(transfer->buffer, count);
	}
	transfer->bufsize = count;
	transfer->status = TRANSFER_STATUS_NORMAL;
	return 0;
}


int transfermode_squidcache_deinit_server(struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);

	/* XXX */
	pi->ci->data_connection.complete = (server_msg->code == 226);

	ftp_message_setm(server_msg,
			226, "Transfer complete",
			0, 0, -1);
	
	close(pi->ci->data_connection.dataserversock);
	pi->ci->data_connection.dataserversock = -1;

	ASSUME(server_msg != NULL);
	ASSUME(server_msg->code);

	return TRANSFERMODE_OK;
}
