/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#ifndef __JFTPGW_H__
#define __JFTPGW_H__

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STRINGS_H   /* SysV */
#include <strings.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>  /* for UINT_MAX -> OSF does not accept inet_addr() == -1 */

/* These headers don't depend on any of the struct definitions in this
 * header file. There is another include-section for those files which do
 * below */

#include "types.h"
#include "util.h"
#include "socketio.h"
#include "sockets.h"
#include "states.h"
#include "events.h"
#include "auth.h"

/* include the autoheader file config.h */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* #include <dmalloc.h> */

#define MAX_VAL(a,b) ((a)>(b)?(a):(b))
#define MIN_VAL(a,b) ((a)<(b)?(a):(b))

#define SERVERADDR		1
#define CLIENTADDR		2

#define TRANSPARENT_YES		1
#define TRANSPARENT_NO		2

#define UID			1
#define EUID			2
#define GID			3
#define EGID			4

#define PRIV			1
#define UNPRIV			2

#define WHITESPACES		" \t"

#ifndef INPORT_ANY
#define INPORT_ANY		0
#endif

#define SVR_LAUNCH_CMDLINE	0
#define SVR_LAUNCH_LOGFILES	1
#define SVR_LAUNCH_READY	2


#if !defined(HAVE_SNPRINTF) || !defined(HAVE_VSNPRINTF)
  int snprintf (char *str, size_t count, const char *fmt, ...);
  int vsnprintf (char *str, size_t count, const char *fmt, va_list arg);
#endif


extern FILE* logfile;
struct allowed_lst;

struct protocolhandler;
struct cmdlogtarget;

struct serverinfo {
	struct {
		int serversocket;
		int clientsocket;
		unsigned long int addr_to_server;
		unsigned long int addr_to_client;
		unsigned long int server_ip;
		char* server_name; /* XXX */
		unsigned long int client_ip;
		char* client_name; /* XXX */
		unsigned long int proxy_ip;
		unsigned int proxy_port;
		int transparent;
		struct sockaddr_in transparent_destination;
		int* boundsocket_list;
		int boundsocket_niface;  /* number of ifaces we bound to */
		struct connliststruct* connected_clients;
	} connection;
	int tcp_wrapper;
	int multithread;
	int servertype; /* SERVERTYPE_(STANDALONE|INETD) */
	int main_server_pid;
	int pidfile_created;
	int chrooted;
	int ready_to_serve;
	/* We need this to re-read the configuration file. If we destroy the
	 * configuration file, we don't know anymore what our chroot-path
	 * was and can't strip the chroot-path to reread the file. */
	char* chrootdir_saved;
	char* conffilename;
	char* binaryname;
	float throughput;  /* XXX */
	float throughput_send;
	float throughput_receive;
	struct hostent_list* hostcache;
	struct uidstruct runasuser;
};

struct proxyinfo {
	struct clientinfo* ci;
	struct serverinfo* si;
	struct protocolhandler* protocol;
};

struct protocolhandler {
	const char* proto_name;
	const char* proxy_name;
	int(*proto_reject)(struct proxyinfo*);
	int(*proto_connect)(struct proxyinfo*);
	int(*proto_accept)(struct proxyinfo*);
	void(*proto_log_cmd_to_target)(struct cmdlogtarget*, struct proxyinfo*);
	char*(*proto_log_replace_char)(const char, struct proxyinfo*);
	int(*proto_allocate)(struct proxyinfo*);
	int(*proto_deallocate)(struct proxyinfo*);
	int(*proto_init_modules)(void);
	int(*proto_config_install)(void);
	int(*proto_read_config_before)(struct proxyinfo*);
	int(*proto_read_config_after)(struct proxyinfo*);
	char* defaultconfigfile;
};

struct transfer_t {
	char* buffer;
	int bufsize;
	int status;
};

/* transfer_t.status may be one of those */

#define TRANSFER_STATUS_NORMAL		0
#define TRANSFER_STATUS_FINISHED	1
#define TRANSFER_STATUS_ERROR		2 /* XXX ? */



/* the following header files depends on some struct definitions, that's why
 * it goes here */
#include "config_header.h"
#include "log.h"
#include "filter.h" /* struct transfer_t */

/* main entry function for all service proxies */
int jproxygw_main(int, char**, struct protocolhandler);

/* signal handling functions */
void read_default_conf(int);
void terminate (int);

int reread_config(struct proxyinfo*);

/* atexit functions */
void sayterminating(void);

/* clean up functions */
void closedescriptors(struct proxyinfo*);
int removepidfile(struct proxyinfo*);

int waitclient(const char*, struct proxyinfo*, struct protocolhandler*);
int inetd_connected(int, struct proxyinfo*, struct protocolhandler*);
int set_userdest(const char*, struct proxyinfo*, const char*);
int passcmd(const char*, struct proxyinfo*);

int daemonize(void);
int stage_action(const char* stage, struct proxyinfo*);

/* from config.c */
int passcmd_check(const char*);


const char* strip_chrooted_path(const char*);


/* from rel2abs.c */
char* rel2abs(const char* path, const char* base,
			char* result, const size_t size);


#endif /* __JFTPGW_H__ */
