/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <sys/time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include "jproxygw.h"


#ifdef HAVE_CRYPT_H
#include <pwd.h>    /* getpass() */
#define _XOPEN_SOURCE
#include <crypt.h>  /* crypt() */
#endif

extern int timeout;
extern int chlds_exited;

void enough_mem(const void* ptr) {
	if (ptr == (void*) 0) {
		jlog(0, "Not enough memory for malloc. Exiting.");
		exit(1);
	}
}

/* concating snprintf
 *  *
 *  * determines the length of the string pointed to by `os', appending
 *  * formatted string to a maximium length of `len'.
 *  *
 *  */

void scnprintf (char *os, size_t len, const char *str, ...) {
	va_list vl;
	char *ostmp = os + strlen (os);

	va_start (vl, str);
	vsnprintf (ostmp, len - strlen (os) - 1, str, vl);
	va_end (vl);

	return;
}


char* my_strcasestr(const char* haystack, const char* needle) {
	char* nhay = strdup(haystack);
	char* nneed = strdup(needle);
	const char* match;
	char* p;

	enough_mem(nhay);
	enough_mem(nneed);

	p = nhay;
	while (*p) {
		*p = (char) toupper((int)*p);
		p++;
	}
	p = nneed;
	while (*p) {
		*p = (char) toupper((int)*p);
		p++;
	}
	match = strstr(nhay, nneed);

	if (match) {
		match = haystack + (match - nhay);
	}

	free(nhay);
	free(nneed);

	return (char*) match;
}


/* check out with:
 *
 * ps ax -o pid,ppid,uid,gid,suid,euid,egid,sgid,cmd | grep 'jftpgw\|EUID'
 *
 */
int changeid(int who_id, int what, const char* reason, struct proxyinfo* pi) {
	int i = 0;
	char* wstr;
	const char* user = (char*) 0;
	const char* group = (char*) 0;
	const char* s_user = (char*) 0;
	const char* s_group = (char*) 0;
	uid_t uid  = getuid();
	uid_t euid = geteuid();

	uid_t tuid; /* Target ids we want to change to */
	uid_t teuid;
	gid_t tgid;
	gid_t tegid;

	uid_t s_uid;
	gid_t s_gid;

	uid_t value; /* just for the logging statement */

	if (who_id == PRIV) {
		tuid = 0; value = 0;
		wstr = "UID";
		/* can't change the (E)(U/G)ID if we are not root */
		if (uid != 0) {
			return 0;
		}
		/* we are done if we are root already */
		if (euid == 0) {
			return 0;
		}
		i = setuid(tuid);
	} else {
		/* who_id == UNPRIV */
		s_user = pi->si->runasuser.username;
		s_uid = pi->si->runasuser.uid;
		s_group = pi->si->runasuser.groupname;
		s_gid = pi->si->runasuser.gid;
		user = config_get_option("runasuser");
		group = config_get_option("runasgroup");

		if (what == UID || what == EUID) {
			/* no user is set - ok */
			if (!user) {
				return 0;
			}
			/* down here we have a user set */
			if (!s_user || strcmp(user, s_user) != 0) {
				jlog(2, "Invalid user specified: %s", user);
			}
		}
		if (what == GID || what == EGID) {
			/* remember that we might not have a runasgroup setting
			 * but we may have read the groupname as the default
			 * group name for the runasuser setting */

			/* no group is set - ok */
			if (!group && !s_group) {
				return 0;
			}
			/* down here we have a group set */
			if (!s_group) {
				jlog(2, "Invalid user specified: %s", group);
			}
		}
		/* can't change the (E)(U/G)ID if we are not root */
		if (uid != 0) {
			return 0;
		}
		if (euid != 0) {
			/* UID is root, but effective UID is not. The GID
			 * should be changed. Change to root first. */
			if (changeid(PRIV, UID, "Want to change UID/GID. "
				"Changing back to root first", pi) < 0) {
				return -1;
			}
		}
		jlog(7, "%s", reason);
		switch(what) {
			case EUID:
				wstr = "EUID";
				teuid = s_uid; value = teuid;
#ifdef HAVE_SETEUID
				i = seteuid(teuid);
#else
				/* HP-UX does not know seteuid() */
				i = setreuid(-1, euid);
#endif
				break;
			case UID:
				wstr = "UID";
				tuid = s_uid; value = tuid;
				/* we are done if we are root already */
				if (tuid == 0 && euid == 0) {
					return 0;
				}
				i = setuid(tuid);
				break;
			case EGID:
				wstr = "EGID";
				tegid = s_gid; value = tegid;
#ifdef HAVE_SETEGID
				i = setegid(tegid);
#else
				/* HP-UX does not know setegid() */
				i = setregid(-1, tegid);
#endif
				break;
			case GID:
				wstr = "GID";
				tgid = s_gid; value = tgid;
				i = setgid(tgid);
				break;
			default:
				wstr = "ERROR";
				value = 0;
		}
	}

	if (i) {
		jlog(3, "Could not change the %s to %d: %s",
				wstr, value, strerror(errno));
	} else {
		jlog(8, "Changed %s to %d", wstr, value);
	}
	return i;
}


/* just get the status of the child so that it can end */
void reap_chld_info (int signo) {
	int err = errno;
	int status;
	/* signal handler but waitpid() is reentrant */
	while (waitpid (-1, &status, WNOHANG) > 0) {};
	errno = err;
}

/* Checks if PATTERN is the beginning of RESPONSE */
int checkbegin(const char* response, const char* pattern) {
	if (strlen(response) < strlen(pattern)) {
		return 0;
	}
	return !strncasecmp(response, pattern, strlen(pattern));
}

/* extracts the numerical code out of an FTP server response. */
int respcode(const char* response) {
	int resp = 0;
	int i;
	const char* respoff = response;

	if (!response) {
		return 0;
	}

	while (respoff[3] != ' ') {
		respoff = strchr(respoff, '\n');
		if (!respoff) {
			return -1;
		} else {
			respoff++;   /* skip over '\n' */
		}
	}

	i = sscanf(respoff, "%d ", &resp);
	if (i != 1 && resp < 100) {
		return -1;
	}
	return resp;
}

char* strip_respcodes(const char* msg) {
	size_t length;
	char* new;
	const char* p;
	const char* end;

	if (!msg) {
		return (char*) 0;
	}

	new = (char*) malloc(1);
	new[0] = '\0';
	
	length = 0;
	p = msg;
	do {
		if (isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2])
					&& (p[3] == ' ' || p[3] == '-')) {

			p += 4;
		}
		end = strstr(p, "\r\n");
		if (end) {
			length += (end - p) + 2; /*= strlen("\r\n") */
			new = realloc(new, length + 1);
			strncat(new, p, (end - p) + 2);
			end += 2;
			p = end;
		} else {
			if (*p) {
				length += strlen(p);
				new = realloc(new, length + 1);
				strcat(new, p);
			}  /* else strlen(p) == 0 - nothing more to add */

		}
	} while (end);

	return new;
}


/* sets HOWMANY bits on. */

unsigned long int setlastbits(int howmany) {
	unsigned long e = 0;
	int i;

	for (i=0; i < 32; i++) {
		if (i <= howmany) {
			e |= 1;
		}
		e = e << 1;
	}

	return ntohl(e);
}

void toupstr(char* s) {
	while (*s) {
		*s = toupper((int)*s);
		s++;
	}
}

void char_squeeze(char *const s, int c) {
	char last = 0;
	int i = 0, j = 0;
	char* tmp = malloc(strlen(s) + 1);
	enough_mem(tmp);

	do {
		if (s[i] == last && s[i] == c) { /* do nothing */ }
		else { tmp[j++] = s[i]; }

		last = s[i];

		i++;
	} while (s[i]);
	tmp[j] = '\0';
	strcpy(s, tmp);
	free(tmp);
}


struct ip_t parse_ip(const char* s) {
	struct ip_t s_ip;
	struct in_addr iaddr;
	int ret;
	const char* slash = strchr(s, '/');
	const char* dot;
	char* ip;
	if (!slash) {
		/* no netmask specified */
		slash = s + strlen(s);
	}
	ip = malloc(slash - s + 1);
	enough_mem(ip);
	strncpy(ip, s, slash - s);
	ip[slash - s] = '\0';

	ret = inet_aton(ip, &iaddr);
	free(ip); ip =0;
	if (ret == 0) {
		/* inet_aton error */
		s_ip.ip = -1;
		s_ip.netmask = -1;
		return s_ip;
	}
	s_ip.ip = iaddr.s_addr;

	if (*slash) {
		slash++;
	}
	/* slash points to the netmask now or is 0 if none has been specified */
	if (!*slash) {
		s_ip.netmask = -1;   /* 255.255.255.255 */
		return s_ip;
	}
	dot = strchr(slash, '.');
	if (!dot) {
		/* a decimal number netmask has been specified */
		s_ip.netmask = setlastbits(atoi(slash));
		return s_ip;
	}
	ret = inet_aton(slash, &iaddr);
	if (ret == 0) {
		/* error */
		s_ip.netmask = inet_addr("255.255.255.255");
	} else {
		s_ip.netmask = iaddr.s_addr;
	}

	return s_ip;
}

char* extract_path(const char* pathfile) {
	const char* last_slash = strrchr(pathfile, '/');
	char* path;
	size_t size;

	if (!last_slash) {
		/* no path, it's just a filename */
		return strdup(".");
	}

	size = last_slash - pathfile + 1 + 1;

	path = (char*) malloc(size);
	enough_mem(path);

	snprintf(path, size, "%s", pathfile);

	return path;
}

/* XXX merge with base_name() */
char* extract_file(const char* pathfile) {
	const char* last_slash = strrchr(pathfile, '/');
	char* file;
	size_t size;

	if ( ! last_slash) {
		/* no path, it's just a filename */
		return strdup(pathfile);
	}

	size = strlen(pathfile) - ( last_slash - pathfile + 1 ) + 1;

	file = (char*) malloc(size);
	enough_mem(file);

	snprintf(file, size, "%s", last_slash + 1);

	return file;
}

int cryptcmp(const char* encrypted, const char* clear) {
	const char* cmp;
#ifdef HAVE_CRYPT
	cmp = crypt(clear, encrypted);
#else
	cmp = clear;
#endif
	if ( ! clear || ! encrypted ) {
		return 1;
	}
	return strcmp(cmp, encrypted);
}

static char* cryptpw(const char* clear) {
#ifdef HAVE_CRYPT
	char* crypted = 0;
	char* ret = 0;
	char salt[3];
	char saltposs[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
		'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
		'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E',
		'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
		'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0',
		'1', '2', '3', '4', '5', '6', '7', '8', '9', '.', '/' };
	int idx1, idx2;

	time_t tm;
	tm = time(NULL) * getpid();
	srand(tm);

	idx1 = (int)((double)rand() / ((double)RAND_MAX + 1) * sizeof saltposs);
	idx2 = (int)((double)rand() / ((double)RAND_MAX + 1) * sizeof saltposs);
	salt[0] = saltposs[idx1];
	salt[1] = saltposs[idx2];
	salt[2] = '\0';

	crypted = crypt(clear, salt);
	ret = (char*) malloc(strlen(crypted) + 1);
	enough_mem(ret);
	strcpy(ret, crypted);
	printf("%s\n", crypted);
	return ret;
#else
	return "foo";
#endif
}


/* encrypt_password() just reads a password from stdin and outputs the
 * encrypted version */

void encrypt_password() {
#ifdef HAVE_CRYPT
	char* pw = getpass("Password: ");
	char* crypted = cryptpw(pw);
	memset(crypted, 0, strlen(pw));
#else
	char* crypted = "No crypt support compiled in.";
#endif
	printf("%s\n", crypted);
}


char* to_ascii(char *data, int *len, int strictconversion) {
	int count, len2 = *len;
	char* buffer2 = (char*) malloc(*len * 2);
	int last = 0;
	char *b2ptr = buffer2;

	for (count = 0; count < len2; count++) {
		if ((data[count] == 10)) {
			if (strictconversion || last != 13) {
				*b2ptr = 13;
				b2ptr++;
				(*len)++;
			}
		}
		*b2ptr = data[count];
		last = *b2ptr;
		b2ptr++;
	}
	return buffer2;
}

int change_root(const char* stage, struct proxyinfo* pi) {
	int i;
	int change_id_back = 0;
	const char* directory = config_get_option("changerootdir");
	const char* req_stage = config_get_option("changeroot");

	if (pi->si->chrooted) {
		return 0;
	}
	if (strcasecmp(req_stage, stage) != 0) {
		return 0;
	}
	if (!directory) {
		return 0;
	}
	if (geteuid() != 0) {
		if (getuid() != 0) {
			jlog(4, "Not root - cannot chroot(). Maybe you drop privileges too early");
			/* make the error dramatic... */
			return -1;
		}
		/* getuid == 0 but geteuid != 0 */
		if (changeid(PRIV, UID,
				"Changing UID to root for chroot", pi) < 0) {
			return -1;
		}
		change_id_back = 1;
	}
	/* getuid == 0, we have a directory and the state matches */

	i = chdir(directory);
	if (i) {
		jlog(4, "Could not chdir to %s: %s", directory, strerror(errno));
		changeid(UNPRIV, EUID, "After chroot (failed)", pi);
		return -1;
	}
	i = chroot(directory);
	if (!i) {
		jlog(7, "Changed root directory to %s", directory);
		pi->si->chrooted = 1;
	} else {
		jlog(4, "Error changeing root directory to %s: %s",
						directory, strerror(errno));
		changeid(UNPRIV, EUID, "After chroot (failed)", pi);
		return -1;
	}
	if (changeid(UNPRIV, EUID, "After chroot (succeeded)", pi) < 0) {
		return -1;
	}
	i = chdir("/");
	if (i) {
		jlog(4, "Could not chdir to / of the chrooted environment: %s",
							strerror(errno));
		return -1;
	}
	return 0;
}

int dropprivileges(const char* stage, struct proxyinfo* pi) {
	if (config_compare_option("dropprivileges", stage)) {
		/* we might not have a runasgroup setting but we might have
		 * read the default group of the runasuser-setting */
		if (pi->si->runasuser.groupname) {
			if (changeid(UNPRIV, GID,
					"Dropping group privileges", pi) < 0) {
				return -1;
			}
		}
		if (config_get_option("runasuser")) {
			if (changeid(UNPRIV, UID,
					"Dropping user privileges", pi) < 0) {
				return -1;
			}
		} else {
			if (getuid() == 0) {
				/* don't bring up a message if we aren't
				 * root anyway */
				jlog(2, "Option runasuser not set but dropping "
					"privileges requested - terminating");
				jlog(2, "If you don't have a dropprivileges option in your configuration file at all, this is because dropprivileges is a standard option. If you really want to run as root, set \"dropprivileges never\" in the configuration file.");
				return -1;
			}
		}
	}
	return 0;
}


/*
 * Takes a socket descriptor and returns the string contain the peer's
 * IP address.
 */
const char *get_char_peer_ip(int fd) {
	unsigned long int ip;
	struct in_addr addr;

	ip = get_uint_peer_ip(fd);
	addr.s_addr = ip;

	return inet_ntoa(addr);
}


/*
 * Takes a socket descriptor and returns the peer's IP address as a unsigned
 * int
 */
unsigned long int get_uint_peer_ip(int fd) {
	struct sockaddr_in name;
#ifdef HAVE_SOCKLEN_T
	socklen_t namelen;
#else
	int namelen;
#endif
	namelen = sizeof(name);

	if (getpeername(fd, (struct sockaddr *) &name, &namelen) != 0) {
		jlog(2, "Could not get peername: %s", strerror(errno));
		return UINT_MAX;
	}

	return name.sin_addr.s_addr;
}

int get_interface_ip(const char* iface, struct sockaddr_in *sin) {
#ifdef HAVE_SIOCGIFADDR
	struct ifreq req;
	int fd = socket(PF_INET, SOCK_DGRAM, 0); 
	int ret;

#ifndef IF_NAMESIZE
#  define IF_NAMESIZE IFNAMSIZ
#endif

	memset(req.ifr_name, 0, IF_NAMESIZE);
	strncpy(req.ifr_name, iface, IF_NAMESIZE - 1);
	req.ifr_addr.sa_family = AF_INET; 

	ret = ioctl(fd, SIOCGIFADDR, &req);
	close(fd);

	if (ret == 0) {
		memcpy(sin, &req.ifr_addr, sizeof(struct sockaddr_in));
		return 0;
	}
#endif
	return -1;
}

int get_interface_name(const struct sockaddr_in sin_req, char* iface) {
#ifdef HAVE_SIOCGIFCONF
	struct ifconf ifc;
	struct sockaddr_in sin;
	int ret, fd = -1, nr = 30, n, found;
	struct ifreq *ifr;

	fd = socket (PF_INET, SOCK_DGRAM, 0);

	if (fd < 0) {
		jlog(7, "Failed to create socket in get_interface_name");
		return -1;
	}

	memset (&ifc, 0, sizeof(ifc));
	ifc.ifc_buf = (void*) 0;
	ifc.ifc_len = sizeof(struct ifreq) * nr;
	ifc.ifc_buf = malloc(ifc.ifc_len);
	enough_mem(ifc.ifc_buf);

	for (;;) {
		ifc.ifc_len = sizeof(struct ifreq) * nr;
		ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);
		enough_mem(ifc.ifc_buf);

		if ((ret = ioctl(fd, SIOCGIFCONF, &ifc)) < 0) {
			jlog(5, "ioctl error: %s", strerror(errno));
			break;
		}
		if (ifc.ifc_len == sizeof(struct ifreq) * nr) {
			/* assume it overflowed and try again */
			nr += 10;
			continue;
		}
		break;
	}

	if (ret < 0) {
		free(ifc.ifc_buf);
		return -1;
	}

	/* loop through interfaces returned from SIOCGIFCONF */
	found = 0;
	ifr = ifc.ifc_req;
	for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
		if (get_interface_ip(ifr->ifr_name, &sin) == 0) {
			if (sin.sin_addr.s_addr == sin_req.sin_addr.s_addr) {
				snprintf(iface, IF_NAMESIZE, 
							"%s", ifr->ifr_name);
				jlog(8, "Interface %s was IP %s", ifr->ifr_name,
					inet_ntoa(sin.sin_addr));
				found = 1;
				break;
			}
		}
		ifr++;
	}

	/* we don't need this memory any more */
	free (ifc.ifc_buf);
	close (fd);
	if (found == 1) {
		return 0;
	}
#endif
	return -1;
}


const char* conv_ip2char(unsigned long int ip) {
	struct in_addr addr;
	addr.s_addr = ip;
	return inet_ntoa(addr);
}

char* mconv_ip2char(unsigned long int ip) {
	struct in_addr addr;
	char* s;
	addr.s_addr = ip;
	s = strdup(inet_ntoa(addr));
	enough_mem(s);
	return s;
}

void term_before_newline(char* s) {
	char* nl = strstr(s, "\r\n");
	if (!nl) {
		return;
	}
	*nl = '\0';
}

void term_before_last_newline(char* s) {
	if (!s || strlen(s) == 0) {
		return;
	}
	while (strlen(s) > 0 &&
		(s[strlen(s) - 1] == '\r' || s[strlen(s) - 1] == '\n')) {

		s[strlen(s) - 1] = '\0';
	}
}

void replace_not_larger(char* s, char* replace_what, char* replace_with) {
	char* p, *pend;
	p = s;

	while ((p = strstr(p, replace_what))) {
		pend = p + strlen(replace_what);
		snprintf(p, strlen(p)+1, "%s%s", replace_with, pend);
	}
}

char* replace_larger(char* s, char* replace_what, char* replace_with) {
	char* new, *start = s;
	int count = 0;
	int size;
	int dst_offset, src_offset;
	char* old_start;

	if (!s) {
		return (char*) 0;
	}
	if (!replace_what || !replace_with
				|| !replace_what[0] || !replace_with[0]) {
		new = strdup(s);
		enough_mem(new);
		return new;
	}
	while ((start = strstr(start, replace_what))) {
		count++;
		start += strlen(replace_what);
	}
	size = strlen(s) - count*strlen(replace_what) + count*strlen(replace_with);
	new = (char*) malloc(size + 1);
	enough_mem(new);

	dst_offset = 0;
	src_offset = 0;
	old_start = s;
	while ((start = strstr(old_start, replace_what))) {
		strncpy(new + dst_offset, s + src_offset, start - old_start);
		dst_offset += start - old_start;
		src_offset += start - old_start;
		strncpy(new + dst_offset, replace_with, strlen(replace_with));
		dst_offset += strlen(replace_with);
		src_offset += strlen(replace_what);
		old_start = start + strlen(replace_what);
	}
	strncpy(new + dst_offset, s + src_offset, strlen(s + src_offset) + 1);
	return new;
}

char* char_prepend(const char* prefix, const char* ostr) {
	char* nstr;
	int newsize;

	if ( ! ostr && ! prefix ) {
		return (char*) 0;
	}
	if ( ! ostr ) {
		nstr = strdup(prefix);
		enough_mem(nstr);
		return nstr;
	}
	if ( ! prefix ) {
		nstr = strdup(ostr);
		enough_mem(nstr);
		return nstr;
	}
	newsize = strlen(ostr) + strlen(prefix) + 1;
	nstr = (char*) malloc( newsize );
	enough_mem(nstr);
	snprintf(nstr, newsize, "%s%s", prefix, ostr);

	return nstr;
}

char* char_append(const char* ostr, const char* suffix) {
	return char_prepend(ostr, suffix);
}

char* char_enclose(const char* prefix, const char* ostr, const char* suffix) {
	char* nstr1, *nstr2;

	nstr1 = char_prepend(prefix, ostr);
	nstr2 = char_prepend(nstr1, suffix);

	free(nstr1);

	return nstr2;
}

char* strnulldup(const char* str) {
	char* s;
	if ( ! str ) {
		return (char*) 0;
	}

	s = strdup(str);
	enough_mem(s);

	return s;
}

char* prependcode(const char* s, int code, int is_last_msg) {
	char* sdup = strdup(s);
	char* nextline = sdup;
	char* buffer;
	char* end, *replaceend, *bufferend, *bufferstart;
	char dashp = '-';
	const char* const linefeed = "\r\n";
	const char* const empty = "";
	const char* append = linefeed;
	int count = 0;
	int i, length;
	size_t bufsize;

	enough_mem(sdup);
	replace_not_larger(sdup, "\\r", "\r");
	replace_not_larger(sdup, "\\n", "\n");

	if (code < 0) {
		code = -code;
	}
	if (code > 999) {
		code = code % 1000;
	}
	if (code < 100) {
		code += 100;
	}

	/* count the newlines */
	while ((nextline = strstr(nextline, "\r\n"))) {
		count++;
		nextline++;  /* thus this position does not match again */
	}

	/* memory: (count + 1) * strlen("xxx ") + 4 (to terminate for sure
	 * with "\r\n") + 1 (Terminator)
	 *
	 * memory = (count + 1) * 4  + 1
	 */

	bufsize = strlen(sdup) + (count + 1) * 4  + 4  + 1;
	bufferstart = buffer = (char*) malloc(bufsize);
	enough_mem(buffer);
	bufferend = buffer + bufsize - 1;

	end = nextline = sdup;
	while (strlen(end)) {
		end = strstr(nextline, "\r\n");
		if (!end) {
			end = nextline + strlen(nextline);
			replaceend = end;
			dashp = ' ';
			append = linefeed;
		} else {
			/* don't replace the first '\' */
			replaceend = end - 1; 
			end += strlen("\r\n");
			if (end == nextline + strlen(nextline)) {
				/* at the very end */
				dashp = ' ';
			} else {
				dashp = '-';
			}
			append = empty;
		}
		i = 0;
		while ((nextline + i) < replaceend) {
			if (nextline[i] == '%') {
				/* Remove the '%' */
				nextline[i] = '/';
			}
			/* remove any special characters between nextline
			 * and end */
			if ((unsigned char) nextline[i] < 32) {
				nextline[i] = '_';
			}
			i++;
		}
		length = MIN_VAL( bufferend - bufferstart,
				  strlen("xxx ")
				  + (end - nextline)
				  + strlen(append)
				  + 1
				);
		if (!is_last_msg) {
			/* do not terminate the string */
			dashp = '-';
			append = linefeed;
		}
		snprintf(bufferstart, length, "%d%c%s%s",
				code, dashp, nextline, append);
		bufferstart += strlen(bufferstart);
		if (!*nextline) { /* end */
			nextline = 0;
		}
		nextline = strstr(nextline, "\r\n") + strlen("\r\n");
	}
	free(sdup);
	return buffer;
}


#define NUMBER_BUFFER   30
char* mconv_int2char(signed int i) {
	char* s = (char*) malloc( NUMBER_BUFFER );
	enough_mem(s);
	snprintf(s, NUMBER_BUFFER, "%d", i);
	return s;
}

/* XXX */
char* mconv_uint2char(unsigned int i) {
	return mconv_int2char(i);
}

char* mconv_lint2char(signed long int i) {
	char* s = (char*) malloc( NUMBER_BUFFER );
	enough_mem(s);
	snprintf(s, NUMBER_BUFFER, "%ld", i);
	return s;
}

char* mconv_luint2char(unsigned long int i) {
	return mconv_lint2char(i);
}

char* mconv_float2char(float f) {
	char* s = (char*) malloc( NUMBER_BUFFER );
	enough_mem(s);
	snprintf(s, NUMBER_BUFFER, "%.2f", f);
	return s;
}

const char* base_name(const char* s) {
	const char* r, *t;

	if (!s) {
		return "(null)";
	}
	t = r = s;
	/* XXX strrchr ? */
	while ((t = strchr(t, '/'))) {
		t++;
		r = t;
	}
	return r;
}

int check_say_timeout(int fd, char* file, int line) {
	if (timeout) {
		jlog(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
		err_time_readline(fd);
		return 1;
	} else {
		return 0;
	}
}

void free_and_zero_char(char** sp) {
	if (sp && *sp) {
		free(*sp);
		*sp = (char*) 0;
	}
}

void check_assumption(int x, const char* xstr, const char* file, int line) {
	if (x) {
		return;
	}
	jlog(1, "FATAL: Assumption %s in file %s, line %d does not hold",
				xstr, file, line);

	abort();
}

char* escape_username(const char* unesc) {
	char* esc;
	const char* p;
	int i;

	if (!unesc) {
		return (char*) 0;
	}
	if (!strchr(unesc, '@') && !strchr(unesc, '"')) {
		esc = strdup(unesc);
		enough_mem(esc);
		return esc;
	}
	
	/* calculate space */
	for (i = 0, p = unesc; *p; p++) {
		if (*p == '"') {
			i++;
		}
	}

	i += 2; /* " at each side */
	i += 1; /* terminator */
	i += strlen(unesc);

	esc = (char*) malloc(i);
	enough_mem(esc);

	esc[0] = '"';
	for (i = 1, p = unesc; *p; p++) {
		if (*p == '"') {
			esc[i] = '\\';
			i++;
		}
		esc[i] = *p;
		i++;
	}
	esc[i] = '"'; i++;
	esc[i] = '\0';

	return esc;
}

char* unescape_username(const char* esc) {
	char* unesc;

	if (!esc) {
		return (char*) 0;
	}
	/* we accept it only as an escaped char if it has " characters at BOTH
	 * ends! */
	if (esc[0] != '"' || esc[strlen(esc)-1] != '"') {
		unesc = strdup(esc);
		enough_mem(unesc);
		return unesc;
	}
	/* actually the new string will be forcibly smaller than the original,
	 * so we don't really need the plus 1 for the terminator... */
	unesc = (char*) malloc(strlen(esc) + 1);
	enough_mem(unesc);

	/* remove outer " */
	strcpy(unesc, esc + 1);
	unesc[strlen(unesc) - 1] = '\0';

	replace_not_larger(unesc, "\\\"", "\"");

	return unesc;
}

void overwrite_string(char* s) {
	memset(s, 0, strlen(s));
}
